/*
 * Normal mappings of chips in physical memory
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/config.h>

#ifdef CONFIG_MTD_PARTITIONS
#include <linux/mtd/partitions.h>
#endif

#define WINDOW_ADDR 0xbe000000
#define WINDOW_SIZE CONFIG_MTD_AR531X_SIZE

/* These ought to be somewhere common... */

#define AR531X_FLASHCTL	0x18400000

#define FLASHCTL_MW	0x30000000      /* Memory width */
#define FLASHCTL_MWx8	0x00000000      /* Memory width x8 */
#define FLASHCTL_MWx16	0x10000000      /* Memory width x16 */
#define FLASHCTL_MWx32	0x20000000      /* Memory width x32 (not supported) */

#define sysRegRead(phys) __raw_readl(KSEG1ADDR(phys))

static struct mtd_info *mymtd;

__u8 ar531x_read8(struct map_info *map, unsigned long ofs)
{
	return __raw_readb(map->map_priv_1 + ofs);
}

__u16 ar531x_read16(struct map_info *map, unsigned long ofs)
{
	return __raw_readw(map->map_priv_1 + ofs);
}

__u32 ar531x_read32(struct map_info *map, unsigned long ofs)
{
	return __raw_readl(map->map_priv_1 + ofs);
}

void ar531x_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

void ar531x_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	__raw_writeb(d, map->map_priv_1 + adr);
	mb();
}

void ar531x_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, map->map_priv_1 + adr);
	mb();
}

void ar531x_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	__raw_writel(d, map->map_priv_1 + adr);
	mb();
}

void ar531x_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio(map->map_priv_1 + to, from, len);
}

struct map_info ar531x_map = {
	name: "Physically mapped flash",
	size: WINDOW_SIZE,
	read8: ar531x_read8,
	read16: ar531x_read16,
	read32: ar531x_read32,
	copy_from: ar531x_copy_from,
	write8: ar531x_write8,
	write16: ar531x_write16,
	write32: ar531x_write32,
	copy_to: ar531x_copy_to
};

#ifdef CONFIG_MTD_PARTITIONS
#ifdef CONFIG_MTD_CMDLINE_PARTS
static struct mtd_partition *mtd_parts = 0;
static int                   mtd_parts_nb = 0;
#else
#if (WINDOW_SIZE == 0x200000)
/******************************************************************
 *
 * 2MB Flash Maps
 */
#ifdef CONFIG_WES900A
/* Flash block size = 64K */
static struct mtd_partition ar531x_partitions[] = {
        {
                name:           "zImage",
                size:           0x90000,		/* 576KB */
                offset:         0,
        }, {
                name:           "rootfs",
                size:           0x150000,		/* 1.3MB */
                offset:         MTDPART_OFS_APPEND,
        }, {
                name:           "config",
                size:           0x10000,		/* 64KB */
                offset:         MTDPART_OFS_APPEND,
        }, {
                name:           "calibration",
                size:           0x10000,		/* 64KB */
                offset:         MTDPART_OFS_APPEND,
        }

};
#else
/* Flash block size = 4K */
static struct mtd_partition ar531x_partitions[] = {
        {
                name:           "zImage",
                size:           0xC0000,		/* 768KB */
                offset:         0,
        }, {
                name:           "rootfs",
                size:           0x130000,		/* 1.3MB */
                offset:         MTDPART_OFS_APPEND,
        }, {
                name:           "config",
                size:           0xf000,			/* 60KB */
                offset:         MTDPART_OFS_APPEND,
        }, {
                name:           "calibration",
                size:           0x1000,			/* 4KB */
                offset:         MTDPART_OFS_APPEND,
        }

};
#endif
#elif (WINDOW_SIZE == 0x400000)
/******************************************************************
 *
 * 4MB Flash Maps
 */
#ifdef CONFIG_USI_AP3X
/* Flash block size = 128K */
static struct mtd_partition ar531x_partitions[] = {
        {
                name:           "zImage",
                size:           0xA0000,		/* 640KB */
                offset:         0,
        }, {
                name:           "rootfs",
                size:           0x280000,		/* 2.6MB */
                offset:         MTDPART_OFS_APPEND,
        }, {
                name:           "config",
                size:           0xC0000,		/* 768KB */
                offset:         MTDPART_OFS_APPEND,
        }, {
                name:           "calibration",
                size:           0x20000,		/* 128KB */
                offset:         MTDPART_OFS_APPEND,
        }

};
#elif defined(CONFIG_WES900A) || \
      defined(CONFIG_WGT624) || \
      defined(CONFIG_DWL2100) || \
      defined(CONFIG_CA8_4) || \
      defined(CONFIG_CA8_5)
/* Flash block size = 64K */
static struct mtd_partition ar531x_partitions[] = {
        {
                name:           "zImage",
                size:           0xA0000,		/* 640KB */
                offset:         0,
        }, {
                name:           "rootfs",
                size:           0x2B0000,		/* 2752KB */
                offset:         MTDPART_OFS_APPEND,
        }, {
                name:           "default",
                size:           0x20000,		/* 128KB */
                offset:         MTDPART_OFS_APPEND,
        }, {
                name:           "config",
                size:           0x80000,		/* 512KB */
                offset:         MTDPART_OFS_APPEND,
        }, {
                name:           "calibration",
                size:           0x10000,		/* 64KB */
                offset:         MTDPART_OFS_APPEND,
        }

};
#else
#error "Must define flash partitioning for your board type"
#endif
#elif (WINDOW_SIZE == 0x800000)
/******************************************************************
 *
 * 8MB Flash Maps
 */
/* Flash block size = 128K */
static struct mtd_partition ar531x_partitions[] = {
        {
                name:           "BootImage",		/* RedBoot/ap30img */
                size:           0xd0000,
                offset:         0,
        }, {
                name:           "redbootimg",
                size:           0x50000,
                offset:         MTDPART_OFS_APPEND,
        }, {
                name:           "ap30img",
                size:           0xd0000,
                offset:         MTDPART_OFS_APPEND,
        }, {
                name:           "zImage",
                size:           0xc0000,
                offset:         MTDPART_OFS_APPEND,
        }, {
                name:           "rootfs",
                size:           0x400000,
                offset:         MTDPART_OFS_APPEND,
        }, {
                name:           "default",
                size:           0x10000,
                offset:         MTDPART_OFS_APPEND,
        }, {
                name:           "config",
                size:           0x100000,
                offset:         0x6c0000,
        }, {
                name:           "RedbootConf",
                size:           0x1000,			/* 4KB */
                offset:         0x7c0000,
        }, {
                name:           "calibration",
                size:           0x10000,		/* 64KB */
                offset:         0x7f0000,
        }

};
#else
#error "Need to define partition mapping for selected flash size"
#endif

#define NUM_PARTITIONS	(sizeof(ar531x_partitions)/sizeof(struct mtd_partition))

#endif
#endif

int __init init_ar531x(void)
{
	static const char *rom_probe_types[] = { "cfi_probe", "jedec_probe", "map_rom", 0 };
	const char **type;
	unsigned int flashctl;

       	printk(KERN_NOTICE "ar531x flash device: %x at %x\n",
	       WINDOW_SIZE, WINDOW_ADDR);

	/* This is nasty, but needed as the new AR2312-01 parts only
	 * have an 8 bit bus to the flash; others have 16
	 */

	flashctl = sysRegRead(AR531X_FLASHCTL) & FLASHCTL_MW;

	if (flashctl == FLASHCTL_MWx8)
		ar531x_map.buswidth = 1;
	else if (flashctl == FLASHCTL_MWx16)
		ar531x_map.buswidth = 2;
	else {
		printk(KERN_ERR "ar531x illegal flash buswidth (%#x)\n",
			flashctl);
		return -ENXIO;
	}

	printk(KERN_NOTICE "ar531x flash buswidth detected as %d\n",
	       ar531x_map.buswidth);

	/*
	 * Must make sure that WINDOW_ADDR is in kseg1 (uncached)
	 */
	ar531x_map.map_priv_1 = (WINDOW_ADDR | KSEG1);

	mymtd = 0;
	type = rom_probe_types;
	for(; !mymtd && *type; type++) {
		mymtd = do_map_probe(*type, &ar531x_map);
	}
	if (mymtd) {
		mymtd->module = THIS_MODULE;

		add_mtd_device(mymtd);
#ifdef CONFIG_MTD_PARTITIONS
#ifdef CONFIG_MTD_CMDLINE_PARTS
		mtd_parts_nb = parse_cmdline_partitions(mymtd, &mtd_parts, 
							"phys");
		if (mtd_parts_nb > 0)
		{
			printk(KERN_NOTICE 
			       "Using command line partition definition\n");
			add_mtd_partitions (mymtd, mtd_parts, mtd_parts_nb);
		}
#else
		if (NUM_PARTITIONS != 0) 
		{
			printk(KERN_NOTICE 
			       "Using ar531x partition definition\n");
			add_mtd_partitions (mymtd, ar531x_partitions, NUM_PARTITIONS);
		}

#endif
#endif
		return 0;
	}
	return -ENXIO;
}

static void __exit cleanup_ar531x(void)
{
	if (mymtd) {
		del_mtd_device(mymtd);
		map_destroy(mymtd);
	}
}

module_init(init_ar531x);
module_exit(cleanup_ar531x);


MODULE_LICENSE("GPL");
MODULE_AUTHOR("Instant802 Networks, Inc.");
MODULE_DESCRIPTION("Atheros AR531x MTD map driver");
