#include <jni.h>
#include "i802.h"

#include "jsyscall.h"

#include <stdio.h>
#include <openssl/ssl.h>
#include <openssl/tls1.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <sys/errno.h>
#include <asm/errno.h>
#include <unistd.h>
#include <string.h>

#include "com_instant802_cldc_net_UnixNativeSocket.h"

#undef DEBUG

#ifdef DEBUG
#define I802_DBG_NONE		0x0000
#define I802_DBG_SETUP		0x0001
#define I802_DBG_TCP		0x0002
#define I802_DBG_UDP		0x0004
#define I802_DBG_READ		0x0008

static int i802_debug_level	= I802_DBG_SETUP | I802_DBG_UDP;

#define I802_DBG(level, args...) \
		do { \
			if (level & i802_debug_level) { \
				fprintf(stderr, ##args); \
			} \
		} while (0)

#else
#define I802_DBG(args...)
#endif

typedef unsigned char u8;

void hostapd_hexdump(const char *title, u8 *buf, size_t len)
{
	size_t i;
	printf("%s - hexdump(len=%d):", title, len);
	for (i = 0; i < len; i++)
		printf(" %02x", buf[i]);
	printf("\n");
}

static inline void set_reuseaddr( int sd ) {
	int val = 1;
	KSETSOCKOPT(sd, SOL_SOCKET, SO_REUSEADDR, &val, sizeof val);
}


static inline int set_bind_to_device( int sd, const char *interface_name ) {
	int err = KSETSOCKOPT(sd, SOL_SOCKET, SO_BINDTODEVICE, interface_name, strlen(interface_name));
	/* Return -errno in case of error */
	return (jint) -err;
}

static inline int set_broadcast( int sd ) {
	int val = 1;
	I802_DBG(I802_DBG_UDP, "setting socket %d to broadcast mode\n", sd);
	int err =  KSETSOCKOPT ( sd, SOL_SOCKET, SO_BROADCAST, &val, sizeof val );
	I802_DBG(I802_DBG_UDP, "set socket %d to broadcast mode - return %d\n", sd, err);
	/* Return -errno in case of error */
	return (jint) -err;
}

/* construct a stream socket state */
JNIEXPORT
jint
Java_com_instant802_cldc_net_UnixNativeSocket_createSocketState(JNIEnv *env, jobject ob) {
	int sd, err;

	I802_DBG(I802_DBG_SETUP, "Creating socket state\n");
	err = KSOCKET (AF_INET, SOCK_STREAM, 0, &sd);
	I802_DBG(I802_DBG_SETUP, "Socket state create returned %d, error %d\n", sd, err);

	if (err == 0)
		return ((jint) sd);
	else
		return ((jint) -1);
}

/* construct a datagram socket state */
JNIEXPORT
jint
Java_com_instant802_cldc_net_UnixNativeSocket_createDatagramSocketState(JNIEnv *env, jobject ob) {
    	int sd, err;

	I802_DBG(I802_DBG_UDP, "Creating datagram socket\n");
	err = KSOCKET (AF_INET, SOCK_DGRAM, 0, &sd);
	I802_DBG(I802_DBG_UDP, "Datagram socket create returned %d, error %d\n", sd, err);

	if (err == 0)
		return ( (jint) sd );
	else
		return ( (jint) -1 );
}

/* construct a unix domain datagram socket state */
JNIEXPORT
jint
Java_com_instant802_cldc_net_UnixNativeSocket_createUnixDomainDatagramSocketState(JNIEnv *env, jobject ob) {
    	int sd, err;

	I802_DBG(I802_DBG_UDP, "Creating unix domain datagram socket\n");
	err = KSOCKET (AF_LOCAL, SOCK_DGRAM, 0, &sd);
	I802_DBG(I802_DBG_UDP, "Unix domain Datagram socket create returned %d, error %d\n", sd, err);

	if (err == 0)
		return ( (jint) sd );
	else
		return ( (jint) -1 );
}


/* Set up a socket to and use the state to return the outcome  */
/* listen(socket_state *state, int port) */
JNIEXPORT
jint
Java_com_instant802_cldc_net_UnixNativeSocket_nativeBind(JNIEnv *env, jobject ob, jint sd, jint port) {
	struct sockaddr_in sa_serv;
	int err;

	set_reuseaddr( sd );

	memset (&sa_serv, '\0', sizeof( struct sockaddr_in ));
	sa_serv.sin_family      = AF_INET;
	sa_serv.sin_addr.s_addr = INADDR_ANY;
	sa_serv.sin_port        = htons (port);          /* Server Port number */

	I802_DBG(I802_DBG_SETUP, "Binding socket %d to port %d\n", sd, port);
	err = KBIND (sd, (struct sockaddr*) &sa_serv, sizeof (sa_serv));
	I802_DBG(I802_DBG_SETUP, "Bind returned %d\n", err);

	return (jint) -err;
}

/* Set up a socket to and use the state to return the outcome  */
/* listen(socket_state *state, int port) */
JNIEXPORT
jint
Java_com_instant802_cldc_net_UnixNativeSocket_nativeLocalBind(JNIEnv *env, jobject ob, jint sd, jstring name) {
	struct sockaddr_un addr;
	int err;
	char *pname;

	// get the socket name out of a java string object into a char*
	pname = (*env)->GetStringUTFChars(env,  name,  NULL);

	// shouldn't need to set reusaddr for local sockets.
	//	set_reuseaddr( sd );

	
	// start off with a blank address structure
	memset (&addr, '\0', sizeof( addr ));

	// fill in the fields of the address structure
	addr.sun_family = AF_LOCAL;
	addr.sun_path[0] = '\0';
	
	// copy the socket pathname into the address structure
	snprintf( addr.sun_path + 1, sizeof( addr.sun_path) - 1, "%s", pname );
	 
	// bind the socket to the address structure
	I802_DBG(I802_DBG_SETUP, "Binding socket %d to local socket %s\n", sd, pname );
	//	hostapd_hexdump( "sockaddr structure",  (u8 *) &addr, sizeof( addr ) );
	err = KBIND (sd, (struct sockaddr*) &addr, sizeof (addr));
	I802_DBG(I802_DBG_SETUP, "Bind returned %d\n", err);

	// free the memory used by the name string
	(*env)->ReleaseStringUTFChars(env, name, pname);

	// return the outcome
	return (jint) -err;
}

/* sendto( state, address, port, buffer, length ) */
JNIEXPORT
jint
Java_com_instant802_cldc_net_UnixNativeSocket_nativeLocalSendTo(JNIEnv *env, jobject ob,
	jint sd, jstring name, jbyteArray buffer, jint len) {
	const unsigned char *pbuf;
	char *pname;
	struct sockaddr_un addr;
	int err, sent;

	pname = (*env)->GetStringUTFChars(env,  name,  NULL);

	// start off with a blank address structure
	memset (&addr, '\0', sizeof( addr ));

	// fill in the fields of the address structure
	addr.sun_family = AF_LOCAL;
	addr.sun_path[0] = '\0';
	
	// copy the socket pathname into the address structure
	snprintf( addr.sun_path + 1, sizeof( addr.sun_path) - 1, "%s", pname );

	/*copy the byte array into a chunk of memory */
	pbuf = (unsigned char *) malloc(len);
	if (pbuf == NULL) {
		return (jint) -ENOMEM;
	}
	(*env)->GetByteArrayRegion(env, buffer,  0, len, pbuf);

	/* finally call sendto to send the memory to the address */

	I802_DBG(I802_DBG_UDP, "Sendto %d bytes on socket %d\n", len, sd);
	err = KSENDTO(sd, pbuf, len, 0, (struct sockaddr *) &addr,
			           sizeof(addr), &sent);
	I802_DBG(I802_DBG_UDP, "Sendto sent %d, error %d\n", sent, err);
	free(pbuf);

	// and release the name structure
	(*env)->ReleaseStringUTFChars(env, name, pname);

	if (err == 0)
		return (jint) sent;
	else
		return (jint) -err;
}


/* sendto( state, address, port, buffer, length ) */
JNIEXPORT
jint
Java_com_instant802_cldc_net_UnixNativeSocket_nativeSendTo(JNIEnv *env, jobject ob,
	jint sd, jstring addr, jint port, jbyteArray buffer, jint len) {
	const unsigned char *pbuf;
	char *paddr;
	struct sockaddr_in sa_dest;
	int err, sent;

	paddr = (*env)->GetStringUTFChars(env,  addr,  NULL);

	/** here, paddr is a char* with the address name **/
	memset (&sa_dest, '\0', sizeof(sa_dest));
	sa_dest.sin_family      = AF_INET;
	sa_dest.sin_addr.s_addr = inet_addr(paddr);
	sa_dest.sin_port        = htons (port);          /* Server Port number */

	/*copy the byte array into a chunk of memory */
	pbuf = (unsigned char *) malloc(len);
	if (pbuf == NULL) {
		return (jint) -ENOMEM;
	}
	(*env)->GetByteArrayRegion(env, buffer,  0, len, pbuf);

	/* finally call sendto to send the memory to the address */

	I802_DBG(I802_DBG_UDP, "Sendto %d bytes on socket %d\n", len, sd);
	err = KSENDTO(sd, pbuf, len, 0, (struct sockaddr *) &sa_dest,
			           sizeof(sa_dest), &sent);
	I802_DBG(I802_DBG_UDP, "Sendto sent %d, error %d\n", sent, err);
	free(pbuf);
	(*env)->ReleaseStringUTFChars(env, addr, paddr);

	if (err == 0)
		return (jint) sent;
	else
		return (jint) -err;
}

/* bindToDevice( socket_state *state, String device_name ) */
JNIEXPORT
jint
Java_com_instant802_cldc_net_UnixNativeSocket_nativeBindToDevice(JNIEnv *env, jobject ob,
	jint sd, jstring device_name ) {

	const char *pdevice_name;
	int err;

	pdevice_name = (*env)->GetStringUTFChars(env, device_name, NULL );

	err = set_bind_to_device(sd, pdevice_name);

	/* clean up */
	(*env)->ReleaseStringUTFChars(env, device_name, pdevice_name);

	return (jint) -err;
}

/* connect( socket_state *state, String address, int port ) */
JNIEXPORT
jint
Java_com_instant802_cldc_net_UnixNativeSocket_nativeConnect(JNIEnv *env, jobject ob,
	jint sd, jstring addr, jint port, jint timeout) {
	const char *paddr;
	struct sockaddr_in sa_serv;
	int err;

	paddr = (*env)->GetStringUTFChars(env,  addr, NULL );

	/** here, paddr is a char* with the address name **/
	memset (&sa_serv, '\0', sizeof(sa_serv));
	sa_serv.sin_family      = AF_INET;
	sa_serv.sin_addr.s_addr = inet_addr(paddr);
	sa_serv.sin_port        = htons (port);          /* Server Port number */

	I802_DBG(I802_DBG_SETUP, "Trying to connect %d\n", sd);
	err = KCONNECT(sd, (struct sockaddr*) &sa_serv, sizeof (sa_serv), (int)timeout);
	I802_DBG(I802_DBG_SETUP, "Connect returned %d\n", err);

	/* clean up */
	(*env)->ReleaseStringUTFChars(env, addr, paddr);

	return (jint) -err;
}

/* Set up a socket to listen - returns -1 or socket descriptor */
/* listen(socket_state *state, int backlog) */
JNIEXPORT
jint
Java_com_instant802_cldc_net_UnixNativeSocket_nativeListen(JNIEnv *env, jobject ob,
	jint sd, jint backlog) {
	int err;

	I802_DBG(I802_DBG_SETUP, "Listening on %d\n", sd);
	err = KLISTEN(sd, backlog);
	I802_DBG(I802_DBG_SETUP, "Listen returned %d\n", err);

	return (jint) -err;
}

/* Set up a socket to broadcast mode */
/* broadcast(socket_state *state ) */
JNIEXPORT
jint
Java_com_instant802_cldc_net_UnixNativeSocket_nativeSetBroadcast(JNIEnv *env, jobject ob, jint sd) {

	return set_broadcast(sd);
}

/** Non-blocking accept. return code for current state is -1 if error, or sd.   New state structure is returned if accept succeeds. */
JNIEXPORT
jint
Java_com_instant802_cldc_net_UnixNativeSocket_nativeAccept(JNIEnv *env, jobject ob, jint sd, jint timeout) {

	int accepted = -1;
	int err;

	I802_DBG(I802_DBG_SETUP, "Accepting on %d (timeout %d)\n", sd, timeout);
	err = KACCEPT(sd, NULL, NULL, timeout, &accepted);
	I802_DBG(I802_DBG_SETUP, "Accept on %d returned %d, error %d\n", sd, accepted, err);

	if( err == 0 ) {
		return accepted;
	} else {
		return (jint) -err;
	}
}

/* return true if the error code is EAGAIN */
JNIEXPORT
jboolean
Java_com_instant802_cldc_net_UnixNativeSocket_isEAGAIN(JNIEnv *env, jobject ob, jint err) {
	return (jboolean)(err == -EAGAIN);
}

/* return true if the last operation 'failed' with EPIPE */
JNIEXPORT
jboolean
Java_com_instant802_cldc_net_UnixNativeSocket_isEPIPE(JNIEnv *env, jobject ob, jint err) {
	return (jboolean)(err == -EPIPE);
}

/* return true if the last operation 'failed' with EINTR */
JNIEXPORT
jboolean
Java_com_instant802_cldc_net_UnixNativeSocket_isEINTR(JNIEnv *env, jobject ob, jint err) {
	return (jboolean)(err == -EINTR);
}

/* return true if the last operation 'failed' with EWOULDBLOCK */
JNIEXPORT
jboolean
Java_com_instant802_cldc_net_UnixNativeSocket_isEWOULDBLOCK(JNIEnv *env, jobject ob, jint err) {
	return (jboolean)(err == -EAGAIN);
}

/* return the return code associated with the last socket operation */
JNIEXPORT
jboolean
Java_com_instant802_cldc_net_UnixNativeSocket_isEINPROGRESS(JNIEnv *env, jobject ob, jint err) {
	return (jboolean)(err == -EINPROGRESS);
}

/* return the return code associated with the last socket operation */
JNIEXPORT
jboolean
Java_com_instant802_cldc_net_UnixNativeSocket_isEALREADY(JNIEnv *env, jobject ob, jint err) {
	return (jboolean)(err == -EALREADY);
}

/* return the return code associated with the last socket operation */
JNIEXPORT
jboolean
Java_com_instant802_cldc_net_UnixNativeSocket_isEADDRNOTAVAIL(JNIEnv *env, jobject ob, jint err) {
	return (jboolean)(err == -EADDRNOTAVAIL);
}


/* return the return code associated with the last socket operation */
JNIEXPORT
jboolean
Java_com_instant802_cldc_net_UnixNativeSocket_isTCPIPERROR(JNIEnv *env, jobject ob, jint err) {
	return (jboolean)((err == -ENETDOWN) ||
			  (err == -EPROTO) ||
			  (err == -EPIPE) ||
			  (err == -ENOPROTOOPT) ||
			  (err == -EHOSTDOWN) ||
			  (err == -ENONET) ||
			  (err == -EHOSTUNREACH) ||
			  (err == -EOPNOTSUPP) ||
			  (err == -ENETUNREACH ));
}

JNIEXPORT
void
Java_com_instant802_cldc_net_UnixNativeSocket_destroy(JNIEnv *env, jobject ob, jint sd) {
    int err;
    I802_DBG(I802_DBG_SETUP, "closing socket %d\n", sd);
	err = KSOCKCLOSE(sd);
    I802_DBG(I802_DBG_SETUP, "socket close for %d returned %d\n", sd, err);
}

JNIEXPORT
jint
Java_com_instant802_cldc_net_UnixNativeSocket_nativeRead(JNIEnv *env, jobject ob, jint sd,
							 jbyteArray buf, jint len, jint timeout) {
	unsigned char *pbuf;
	int err, rd;

	pbuf = (unsigned char *) malloc(len);
	if (pbuf == NULL) {
		return (jint) -ENOMEM;
	}

	I802_DBG(I802_DBG_READ, "Reading up to %d bytes on %d\n", len, sd);
	err = KSOCKREAD(sd, pbuf, len, timeout, &rd);
	I802_DBG(I802_DBG_READ, "Read read %d, error %d\n", rd, err);

	if ((err == 0) && (rd > 0)) {
		(*env)->SetByteArrayRegion(env, buf, 0, rd, pbuf);
	}
	free(pbuf);

	if (err == 0) {
		if (rd == 0)
			return (jint) -EPIPE;
		else
			return (jint) rd;
	} else
		return (jint) -err;
}

/**
 * Write.
 * @return > 0 => # bytes written, 0 => try again, < 0 => error
 * */
JNIEXPORT
jint
Java_com_instant802_cldc_net_UnixNativeSocket_nativeWrite(JNIEnv *env, jobject ob, jint sd, jbyteArray buf, jint len) {
	unsigned char *pbuf;
	int written, err;

	pbuf = (unsigned char *) malloc(len);
	if (pbuf == NULL) {
		return (jint) -ENOMEM;
	}
	(*env)->GetByteArrayRegion(env, buf, 0, len, pbuf);
	I802_DBG(I802_DBG_TCP, "Writing %d bytes on %d\n", len, sd);
	err = KWRITE(sd, pbuf, len, &written);
	I802_DBG(I802_DBG_TCP, "Write wrote %d, error %d\n", written, err);

	free(pbuf);
	if (err == 0)
		return (jint) written;
	else
		return (jint) -err;
}

/**
* The return 64 bit quantity is as follows:
* If the bottom 16 bits threated as a signed 16 bit quantity are >= 0 then
*   		The top 32 bits are IP address of sender
*   		The next 16 bits are port of sender
* 		The bottom 16 bits are (signed0 length of data
* Else if bttom 16 bits ==-1 then the upper 32 bits are the negated errno value
* Alanl
*/
JNIEXPORT
jlong
Java_com_instant802_cldc_net_UnixNativeSocket_nativeReceiveFrom(JNIEnv *env, jobject ob, jint sd,
    			jbyteArray jbuffer,  jint length)
{
	struct sockaddr_in addr;
	int len = sizeof(struct sockaddr_in);
	unsigned long long lres;
	unsigned long ip;
	unsigned short port;
	int res, err;
	unsigned char *pbuf;

	pbuf = (unsigned char *) malloc(length);
	if (pbuf == NULL) {
		return (jint) -ENOMEM;
	}

	I802_DBG(I802_DBG_UDP, "Recvfrom %d bytes on %d\n", length, sd);
	err = KRECVFROM(sd, pbuf, length, 0, (struct sockaddr*)&addr, &len, 0, &res);
	I802_DBG(I802_DBG_UDP, "Recvfrom returned %d, error %d\n", res, err);

	if (err == 0) {
		(*env)->SetByteArrayRegion(env, jbuffer, 0, res, pbuf);
	   	ip = htonl(addr.sin_addr.s_addr);
    		port = htons(addr.sin_port);
		lres = ip;
		lres = lres << 32;
		lres = lres | (((unsigned long long)port) << 16);
		lres = lres | (((unsigned long long)res) & 0xFFFF);
	} else  {
		lres = (((unsigned long long)(-err)) << 32) | 0xffffffff;
	}
	free((void *)pbuf);
	return (jlong) lres;
}

/**
* The return 64 bit quantity is as follows:
* If the bottom 16 bits threated as a signed 16 bit quantity are >= 0 then
*   		The top 32 bits are IP address of sender
*   		The next 16 bits are port of sender
* 		The bottom 16 bits are (signed0 length of data
* Else if bttom 16 bits ==-1 then the upper 32 bits are the negated errno value
* return value contains 
*   length of data
*   negative errno if failure
* RobinB
*/
JNIEXPORT
jlong
Java_com_instant802_cldc_net_UnixNativeSocket_nativeLocalReceiveFrom(JNIEnv *env, jobject ob, jint sd,
    			jbyteArray jbuffer,  jint length)
{
	struct sockaddr_un addr;
	int len = sizeof( addr );
	unsigned long long lres;
	unsigned long ip;
	unsigned short port;
	int res, err;
	unsigned char *pbuf;

	pbuf = (unsigned char *) malloc(length);
	if (pbuf == NULL) {
		return (jint) -ENOMEM;
	}

	I802_DBG(I802_DBG_UDP, "Recvfrom %d bytes on %d\n", length, sd);
	err = KRECVFROM(sd, pbuf, length, 0, (struct sockaddr*)&addr, &len, 0, &res);
	I802_DBG(I802_DBG_UDP, "Recvfrom returned %d, error %d\n", res, err);

	if (err == 0) {
	  (*env)->SetByteArrayRegion(env, jbuffer, 0, res, pbuf);
	  lres = res;
	} else  {
	  lres = -err;
	}
	free((void *)pbuf);
	I802_DBG(I802_DBG_UDP, "returning %d", lres );
	return (jlong) lres;
}
