#include <jni.h>
#include "lock-impl.h"
#include "jthread.h"

#include "i802.h"

#include <stdio.h>
#include <openssl/crypto.h>
#include <openssl/ssl.h>
#include <openssl/tls1.h>

#define CONTEXT_MODE_SERVER 	1
#define CONTEXT_MODE_CLIENT 	2
#define CONTEXT_MODE_PEER   	3	
#define ISSERVER(mode)			(mode & 1)

static int 	sid_ctx = 2;
static int	context_inited = 0;

#undef DEBUG
#ifdef DEBUG
#define I802_DBG(args...)	fprintf(stderr, ##args)
#else
#define I802_DBG(...)
#endif

/*
 * SSL's internal thread safety support
 */
static jmutex*	lock_cs;

static unsigned long i802_ssl_thread_id (void) {
	return (unsigned long) jthread_current();
}

static void i802_ssl_locking_callback(int mode,
				      int type,
				      char *file,
				      int line) {
        if (mode & CRYPTO_LOCK)
                {
		I802_DBG("locking mutex %d\n", type);
		jmutex_lock(&(lock_cs[type]));
		I802_DBG("locked mutex %d\n", type);
                }
        else
                {
		I802_DBG("unlocking mutex %d\n", type);
                jmutex_unlock(&(lock_cs[type]));
                }
}


void i802_ssl_init() {
	int i;

	if (! context_inited) {
 		context_inited = 1;

	 	SSL_load_error_strings();                /* readable error messages */
		SSL_library_init();                      /* initialize library */		
 		// actions_to_seed_PRNG();

		// Initialise the SSL library's mutex callbacks

		lock_cs=malloc(CRYPTO_num_locks() * sizeof(jmutex));
		for (i=0; i<CRYPTO_num_locks(); i++)
			{
			jmutex_initialise(&(lock_cs[i]));
			}

		CRYPTO_set_id_callback((unsigned long (*)())i802_ssl_thread_id);
		CRYPTO_set_locking_callback((void (*)())i802_ssl_locking_callback);
	}	
}
	
JNIEXPORT 
void
	Java_com_instant802_cldc_ssl_SSLSessionContext_printErrors(JNIEnv *env, jobject obj) {
		// ()
		ERR_print_errors_fp(stderr);
	}
	
	
JNIEXPORT 
jint
	Java_com_instant802_cldc_ssl_SSLSessionContext_createContext(JNIEnv *env, jobject obj, jint mode) {
		// (int mode)
		SSL_CTX *pctx;
		SSL_METHOD *pmethod;

		i802_ssl_init();		
		
		if (mode == CONTEXT_MODE_SERVER)
 			pmethod = TLSv1_server_method();
		else if (mode == CONTEXT_MODE_CLIENT)
			pmethod = TLSv1_client_method();
		else
			pmethod = TLSv1_method();
		
		pctx = SSL_CTX_new(pmethod);
		if (ISSERVER(mode))
			SSL_CTX_set_session_id_context(pctx, (unsigned char *)&sid_ctx, sizeof(sid_ctx));
		// The damn SSL demands (quietly) that any SSL_write rerties done because
		// of SSL_ERR_WANTS_XX _must_ use exactly the same buffer (i.e. the same
		// memory pointer) else it throws an SSL_ERROR_SSL _unless this sodding obscure
		// value is set. 
		// This has caused some amount of frsutration over the last few weeks
		// (That is called dry, British understatement)
		SSL_CTX_set_mode(pctx, SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER);
		return (jint)pctx;
	}
	
JNIEXPORT
jint
	Java_com_instant802_cldc_ssl_SSLSessionContext_useRSAPrivateKey(JNIEnv *env, jobject obj, jint ctx, jbyteArray filen, jint filenlen, jint type) {
		SSL_CTX *pctx;
		int i;
		char *fname;
		jint ret;
		
		pctx = (SSL_CTX *)ctx; 
		
		i802_ssl_init();		
		
		fname = (char *) malloc(filenlen + 1);
		for(i = 0; i < filenlen; ++i) {
			fname[i] =  (*env)->GetByteArrayElements(env, filen, 0)[i];
		}
		fname[filenlen] = 0; //  null term "C" strings
		ret =  (jint) (SSL_CTX_use_PrivateKey_file(pctx, fname, type));		
		free(fname);
		return ret;
	}

JNIEXPORT 
jint
	Java_com_instant802_cldc_ssl_SSLSessionContext_useCertificateFile(JNIEnv *env, jobject obj, jint ctx, jbyteArray filen, jint filenlen, jint type) {
		SSL_CTX *pctx;
		int i;
		char *fname;
		jint ret;
		
		pctx = (SSL_CTX *)ctx; 
		
		i802_ssl_init();		
		
		fname = (char *) malloc(filenlen + 1);
		for(i = 0; i < filenlen; ++i) {
			fname[i] =  (*env)->GetByteArrayElements(env, filen, 0)[i];
		}
		fname[filenlen] = 0; //  null term "C" strings
		ret =  (jint) (SSL_CTX_use_certificate_file(pctx, fname, SSL_FILETYPE_PEM));	
		return ret;
	}		
	
JNIEXPORT 
jint
	Java_com_instant802_cldc_ssl_SSLSessionContext_close(JNIEnv *env, jobject obj, jint ctx) {
		SSL_CTX *pctx;
		
		pctx = (SSL_CTX *) ctx; 
		SSL_CTX_free(pctx);
	}			

