/*
 * Created on Oct 27, 2003
 * 
 * To change the template for this generated file go to Window - Preferences -
 * Java - Code Generation - Code and Comments
 */
package de.gyronet.kangaroo.io;

import java.io.*;
import java.util.Hashtable;
import javax.microedition.io.Datagram;
import javax.microedition.io.DatagramConnection;

/**
 * @author alanl
 * 
 * This interface defines the capabilities that a datagram connection must
 * have.
 * <p>
 * Reminder: In common with all the other addressing schemes used for I/O in
 * CLDC, the syntax for datagram addressing is not defined in the CLDC
 * Specification. Syntax definition can only be take place at the profile
 * level. The reason for this is that the datagram interface classes of CLDC
 * can be used for implementing various kinds of datagram protocols. Examples
 * include IP and WDP networks as well as infrared beaming protocols used by
 * various PDAs and other devices. All these protocols use very different
 * addressing mechanisms.
 * <p>
 * In the sample implementation provided as part of the CLDC reference
 * implementation, the following addressing scheme is used for UDP datagrams.
 * <p>
 * The parameter string describing the target of a connection in the CLDC
 * reference implementation takes the following form:
 * 
 * <pre>
 *  {protocol}://[{host}]:[{port}]
 * </pre>
 * 
 * 
 * A datagram connection can be opened in a "client" mode or "server" mode. If
 * the "//{host}" part is missing then the connection is opened as a "server"
 * (by "server", we mean that a client application initiates communication).
 * When the "//{host}" part is specified, the connection is opened as a
 * "client".
 * <p>
 * Examples:
 * <p>
 * A datagram connection for accepting datagrams <br>datagram://:1234
 * <p>
 * A datagram connection for sending to a server: <br>
 * datagram://123.456.789.12:1234
 * <p>
 * 
 * Note that the port number in "server mode" (unspecified host name) is that
 * of the receiving port. The port number in "client mode" (host name
 * specified) is that of the target port. The reply-to port in both cases is
 * never unspecified. In "server mode", the same port number is used for both
 * receiving and sending. In "client mode", the reply-to port is always
 * dynamically allocated.
 * <p>
 * The allocation of datagram objects is done in a more abstract way than in
 * J2SE. This is to allow a single platform to support several different
 * datagram interfaces simultaneously. Datagram objects must be allocated by
 * calling the "newDatagram" method of the DatagramConnection object. The
 * resulting object is defined using another interface type called
 * "javax.microedition.io.Datagram".
 *  
 */
public class DatagramConnectionImpl implements DatagramConnection
{

	InetImpl con;

	DatagramConnectionImpl()
	{
	}

	DatagramConnectionImpl(InetImpl con)
	{
		this.con = con;
	}

	DatagramConnectionImpl open(
		URL targetURL,
		Hashtable params,
		int mode,
		boolean timeouts)
		throws IOException
	{
		con = new InetImpl();
		con.timeout = timeouts ? 60000 : 0;
		con.mode = mode;
		try
		{
			con.create0(false); //false means datagram
											  // connection
			con.connect0(targetURL.addr, targetURL.port);
		}
		catch (IOException e)
		{
			try
			{
				con.close();
			}
			catch (IOException _)
			{
			}
			throw (e);
		}

		con.open = true;
		con.openCounter++;

		return this;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.microedition.io.DatagramConnection#getMaximumLength()
	 */
	public int getMaximumLength() throws IOException
	{
		throw new IOException("DatagramConnectionImpl.getMaximumLength not implemented");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.microedition.io.DatagramConnection#getNominalLength()
	 */
	public int getNominalLength() throws IOException
	{
		throw new IOException("DatagramConnectionImpl.getNominalLength not implemented");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.microedition.io.DatagramConnection#send(javax.microedition.io.Datagram)
	 */
	public void send(Datagram dgram) throws IOException
	{
		con.write(dgram.getData(), 0, dgram.getLength());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.microedition.io.DatagramConnection#receive(javax.microedition.io.Datagram)
	 */
	public void receive(Datagram dgram) throws IOException
	{
		con.readBytes(dgram.getData(), 0, dgram.getLength());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.microedition.io.DatagramConnection#newDatagram(int)
	 */
	public Datagram newDatagram(int size) throws IOException
	{
		byte[] buf = new byte[size];
		return new DatagramObject(this, buf, size);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.microedition.io.DatagramConnection#newDatagram(int,
	 *           java.lang.String)
	 */
	public Datagram newDatagram(int size, String addr) throws IOException
	{
		Datagram dg = newDatagram(size);
		dg.setAddress(addr);
		return dg;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.microedition.io.DatagramConnection#newDatagram(byte[], int)
	 */
	public Datagram newDatagram(byte[] buf, int size) throws IOException
	{
		Datagram dg = new DatagramObject(this, buf, size);
		return dg;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.microedition.io.DatagramConnection#newDatagram(byte[], int,
	 *           java.lang.String)
	 */
	public Datagram newDatagram(byte[] buf, int size, String addr)
		throws IOException
	{
		Datagram dg = newDatagram(buf, size);
		dg.setAddress(addr);
		return dg;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.microedition.io.Connection#close()
	 */
	public void close() throws IOException
	{
		con.close();
	}

}
