/*
 * Created on Nov 1, 2003
 * author: Robin Barooah
 * (C) 2003 Instant802 Networks Inc
 */
package com.instant802.cldc.deviceconfiguration;

import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import com.instant802.cldc.radius.RadiusServer;
import com.instant802.cldc.DataPath.XMLRepresentationException;
import com.instant802.cldc.LSQL.APConfException;
import com.instant802.cldc.LSQL.ConfigurationChangeListener;
import com.instant802.cldc.LSQL.LSQLException;
import com.instant802.cldc.channelplanner.ChannelPlanner;
import com.instant802.cldc.messaging.Message;
import com.instant802.cldc.net.MalformedHTTPMessageException;
import com.instant802.cldc.net.SocketException;
import com.instant802.cldc.orchestrator.OrchestratorNode;
import com.instant802.cldc.synchronization.SynchronizationSetMessage;
import com.instant802.cldc.util.Log;

/**
 * @author robin
 */
public final class ConfigurationModel
	extends IndexedConfiguration
	implements ConfigurationChangeListener, MutableConfiguration, ConfigurationSource {

	private static final boolean SHARED_ONLY = false;
	private static final boolean SHARED_AND_LOCAL = true;
	final OrchestratorNode node;
	private RadiusServer radiusServer;
	private ChannelPlanner channelPlanner;
	
	public ConfigurationModel(OrchestratorNode node) {
		this.node = node;
	}

	/** when started, the configuration model should read a copy of the current system configuration into a DOM tree **/
	public synchronized void populate()
		throws
			XMLRepresentationException,
			IOException, MalformedHTTPMessageException, SocketException {
		byte[] doc = node.lsqlClient().configAsDocument();		
		ConfigDOM dom = new ConfigDOM( node.schema(), doc );			
		createObjects( dom );		
		exportState();
	}

	/**
	 * given an XML DOM, clear and repopulate the complete
	 * configuration model.
	 */
	private synchronized void createObjects(ConfigDOM dom )  {
		types.clear();
		Enumeration e = dom.objects();
		while( e.hasMoreElements() ) {
			ConfigObject object = (ConfigObject) e.nextElement();
			addObject( object );
		}
	}

	public synchronized void addObject( ConfigObjectAdapter toAdd )  {
		addObject( toAdd.getObject() );
	}

	private void exportState() throws ElementNotFoundException {
		String clusterId = getOrchestratorConfig().getClusterId();
		node.masterCriteria().setClusterId( clusterId );
	}

	public void notifyRadiusServer() {
		radiusServer.handleConfigChanged(this);
	}
	
	public void notifyChannelPlanner() {
		try
		{
			this.channelPlanner.handleConfigChanged(getChannelPlannerConfig());
		} catch (XMLRepresentationException e)
		{
			if (Log.DEBUG) Log.error("notifyChannelPlanner: threw Exception " + e.toString());
		}
	}

	public synchronized OrchestratorConfig getOrchestratorConfig()
		throws ElementNotFoundException {
			OrchestratorConfig config = new OrchestratorConfig( node.masterCriteria(), getSingleton( OrchestratorConfig.PK ) );
		return config;
	}

	public synchronized DebugConfig getDebugConfig()
		throws ElementNotFoundException {
			DebugConfig config = new DebugConfig( getSingleton( DebugConfig.PK ) );
		return config;
	}

	public synchronized SystemConfig getHostConfig()
		throws ElementNotFoundException {
		SystemConfig hostConfig = new SystemConfig( getSingleton( SystemConfig.PK ) );
		return hostConfig;
	}

	private ConfigObject getSingleton(PKSingleton key) throws ElementNotFoundException {
		Hashtable typeTable = getTypeTable( key.value() );
		ConfigObject object = (ConfigObject) typeTable.get( key.value() );
		if( null == object ) {
			throw new ElementNotFoundException( "couldn't find singleton "+key.value() );
		}
		return object;
	}

	public synchronized ObjectList currentObjects()  {
		IndexedConfiguration list = new IndexedConfiguration();
		Enumeration tables = types.elements();
		while( tables.hasMoreElements() ) {
			Hashtable typeTable = (Hashtable) tables.nextElement();
			Enumeration types = typeTable.elements();
			while( types.hasMoreElements() ) {
				ConfigObject co =  (ConfigObject) types.nextElement() ;
				if (co.isShared())
					list.addObject(co);
			}
		}
		return list;
	}

	/**
	 * update this configuration model so that it matches the supplied list of objects
	 */
	public synchronized void updateFromObjectList(ObjectList to)
		throws LSQLException, APConfException {
		ObjectList from = this.currentObjects();		
		ConfigurationChanges changes = listDifferences(from, to, SHARED_ONLY);
		changes.apply(this);
	}

	/**
	 * Return the set of changes that, when applied to the supplied original configuration
	 * would result in this configuration model.
	 * @param original the original configuration from which to determine the differences
	 * @return the set of changes that could be applied to the original to yield the current config
	 */
	public synchronized ConfigurationChanges changesFrom(ConfigurationModel original)  {
		ObjectList from = original.currentObjects();
		ObjectList to = this.currentObjects();
		ConfigurationChanges changes =
			listDifferences(from, to, SHARED_AND_LOCAL);
		return changes;
	}

	private synchronized ConfigurationChanges listDifferences(
		ObjectList from,
		ObjectList to,
		boolean local)  {
		ConfigurationChanges changes = new ConfigurationChanges( node );
		updatesAndDeletions(from, to, changes, local);
		additions(from, to, changes, local);
		return changes;
	}

	private synchronized void additions(
		ObjectList from,
		ObjectList to,
		ConfigurationChanges changes,
		boolean local) {
		Enumeration f = to.elements();
		while (f.hasMoreElements()) {
			ConfigObject candidate = (ConfigObject) f.nextElement();
			if (include(candidate, local)) {
				if (!from.defines(candidate)) {
					// if the end contains something that the original does not, then we would need to add it to the original
					changes.add(candidate);
				}
			}
		}
	}

	private final boolean include(ConfigObject candidate, boolean local) {
		return node.schema().isShared( candidate ) || (local && node.schema().writeLocally( candidate ) );
	}

	private synchronized void updatesAndDeletions(
		ObjectList from,
		ObjectList to,
		ConfigurationChanges changes,
		boolean local) {
		Enumeration e = from.elements();
		while (e.hasMoreElements()) {
			ConfigObject candidate = (ConfigObject) e.nextElement();
			if (include(candidate, local)) {
				if (!to.defines(candidate)) {
					// if the object isn't defined in the end result set, then we would need to delete it from the original
					if (Log.DEBUG) debug("incoming configuration does not define "+candidate.primaryKey().value( candidate ) +" so marking it for deletion.");
					changes.delete(candidate);
				} else {
					ConfigObject desiredState = to.definition(candidate);
					FieldChanges2 differences = desiredState.differencesFrom( candidate, local );
					if( differences.count() > 0 ) {
						if (Log.DEBUG) debug("incoming configuration has differences in fields "+differences+" in object "+candidate.primaryKey().value( candidate ) );
						if (Log.DEBUG) debug("desired state is: "+desiredState.asXMLDocument() );
						if (Log.DEBUG) debug("current state is: "+candidate.asXMLDocument() );
						changes.update( differences );
					}
				}
			}
		}
	}
	
	private void debug( String message ) {
		Log.debug( message );
	}

	public synchronized void setRadiusServer(RadiusServer server) {
		this.radiusServer = server;

	}

	public void setChannelPlanner(ChannelPlanner cp) {
		this.channelPlanner = cp;
	}
	
	public synchronized Enumeration getClusteredAPs() 
	{
		Hashtable table = getTypeTable( ClusteredAP.TYPE );
		Vector found = new Vector();
		if( table != null ) 
		{
			Enumeration aps = table.elements();
			while( aps.hasMoreElements() ) {
				ClusteredAP ap = new ClusteredAP( (ConfigObject) aps.nextElement() );
				found.addElement( ap );
			}
		}
		return found.elements();
	}

	//TODO: this should probably be done with an enumeration like the ClusteredAPEnumeration
	public synchronized Hashtable getWirelessUsers() {
		Hashtable table = getTypeTable( WirelessUser.TYPE );
		Hashtable found = new Hashtable();
		if( table != null) {
			Enumeration users = table.elements();
			while( users.hasMoreElements() ) {
				WirelessUser user = new WirelessUser( (ConfigObject) users.nextElement() );
				found.put( user.username(), user );
			}
		}
		return found;
	}

	public synchronized ConfigObject getChannelPlannerConfig() {
		Hashtable table = getTypeTable( ChannelPlanner.TYPE );		
		return (ConfigObject) table.get(ChannelPlanner.TYPE);
	}
	
	public Schema schema() {
		return node.schema();
	}

	public synchronized void handleConfigurationChange() throws LSQLException, APConfException {
		try {
			node.lsqlMessageReceiver().setLocallyModifiedIfClustered();			
			final boolean clusterable = getOrchestratorConfig().clusterable();
			populate();
			patch();
			notifyRadiusServer();
			notifyChannelPlanner();
			distributeConfiguration();
			if( ! clusterable ) {
				if( getOrchestratorConfig().clusterable() ) {
					node.exitStandalone(); 
				}
			}
		} catch (XMLRepresentationException e) {
			warning(e);
		} catch (IOException e) {
			warning(e);
		} catch (MalformedHTTPMessageException e) {
			warning(e);
		} catch (SocketException e) {
			warning(e);
		} catch (Exception e) {
			Log.error(e);
		}
	}

	/**
	 * patch up the configuration file to make changes that only orchestrator can do
	 * based on changes made by the UI
	 */
	public synchronized boolean patch() throws LSQLException, APConfException, XMLRepresentationException, IOException, MalformedHTTPMessageException, SocketException {
		if( patchLocation() ) {
			writeToAPConf();
			return true;
		}
		return false;
	}

	/**
	 * check whether the location field in the clusteredAP record matches the one in the orchestrator
	 * object, and update the clusteredAP record if necessary.
	 */
	private synchronized boolean patchLocation() throws ElementNotFoundException {
		String location = getOrchestratorConfig().location();
		Enumeration e = getClusteredAPs();
		String ourId = node.getNodeId();
		
		/* logic here is:
		 * 1. get the ap from the configuration model that matches our ID.
		 * 2. if its location differs from the 'global' one in the configuration file, then update it.
		 * 3. return true if updated so that this change can be propagated to the cluster.
		 */
		
		while( e.hasMoreElements() ) {
			ClusteredAP ap = (ClusteredAP) e.nextElement();
			if( ap.id().equals( ourId ) ) {
				if( ! location.equals( ap.location() ) ) {
					ap.location( location );
					return true;
				}
			}
		}
		return false;
	}

	public synchronized void distributeConfiguration()
		throws ElementNotFoundException, XMLRepresentationException, IOException, MalformedHTTPMessageException, LSQLException, APConfException, SocketException {
		node.coordinationAgent().broadcast(asSynchronizationSet());
	}

	private void warning(Exception e) {
		Log.warning(e);
	}

	public synchronized void writeToAPConf()
		throws
			LSQLException,
			APConfException, XMLRepresentationException, IOException, MalformedHTTPMessageException, SocketException {
		ConfigurationModel actual = new ConfigurationModel(node);
		actual.populate();
		//				debug("current state from apconfd is: ");
		//				actual.debugDump();
		//				debug("current state of configuration model is: ");
		//				debugDump();
		ConfigurationChanges changes = this.changesFrom(actual);
		changes.apply(node.lsqlClient());
		node.lsqlClient().setLocalRunningConfig();
		if (Log.DEBUG) Log.info("reading new state back from apconfd");
		populate();
		if (Log.DEBUG) Log.info("done applying changes to apconfd");
	}

	public synchronized Message asSynchronizationSet()
		throws ElementNotFoundException  {
		return new SynchronizationSetMessage(node, currentObjects());
	}

	public synchronized void updateObject(FieldChanges2 changes)
		throws LSQLException, APConfException {
		
		ConfigObject found = findDefinition( changes.object() );
		if( null != found ) {
			found.updateFrom( changes );
		}
	}

	public void deleteObject(ConfigObjectAdapter adapter) throws LSQLException, APConfException {
		deleteObject( adapter.getObject() );
	}

	public void handleObjectChange(ConfigObject object) {
		node.objectChangeProcessor().handleObjectChange( object );
	}
}
