package com.instant802.cldc.radius;

import com.instant802.cldc.ssl.OpenSSL;
import com.instant802.cldc.net.Datagram;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

/**
 * @author alan
 *
 * To change this generated comment edit the template variable "typecomment":
 * Window>Preferences>Java>Templates.
 * To enable and disable the creation of type comments go to
 * Window>Preferences>Java>Code Generation.
 */

public class RadiusPacket
{

	byte[] data;
	int port;
	String address;
	Hashtable attributes;
	int msgauthindex = 0;
	Datagram p;

	public static final int R_INVALID = 0, /* Internal use */
	R_REQUEST = 1, /* */
	R_ACCEPT = 2, /* */
	R_REJECT = 3, /* */
	R_CHALLENGE = 11, /* */
	R_REAUTHENTICATED = 20; /* Internal use */

	public static final int RADIUS_ATTR_USER_NAME = 1,
		RADIUS_ATTR_STATE = 24,
		RADIUS_ATTR_VENDOR_SPECIFIC = 26,
		RADIUS_ATTR_CALLING_STATION_ID = 31;

	public static final byte MS_MPPE_SEND_KEY = 16, MS_MPPE_RECV_KEY = 17;

	public RadiusPacket(Datagram p)
	{
		data = p.getData();
		this.p = p;
		address = p.getAddress();
		attributes = null;
	}

	public int getCode()
	{
		return data[0];
	}

	public int getLength()
	{
		return getInt(2);
	}

	public int getIdentifier()
	{
		return data[1];
	}

	/**
	 * Build hash table of attribute/(multi)value pairs from the
	 * Radius packet
	 */
	private void initAttributes()
	{
		if (null == attributes)
		{
			int length, pos;
			attributes = new Hashtable();
			length = getInt(2);
			pos = 20; // start of attributes
			// Since attributes have to be order preserving multi-values we have to do ...
			while (pos < length)
			{
				Integer type = new Integer((int) data[pos]);
				int attrlen = (int) (0x00ff & data[pos + 1]);
				byte[] val = new byte[attrlen - 2];
				System.arraycopy(data, pos + 2, val, 0, attrlen - 2);
				// attrlen includes type and length	
				Vector entry;
				if (null == (entry = (Vector) attributes.get(type)))
				{
					entry = new Vector();
					attributes.put(type, entry);
				}
				// Record the position of the MSG-AUTHENTICATOR if it exists 
				// - this saves us a lot of work
				// dorking with copies when trying to verify authenticity
				if (type.intValue() == EAPHandler.ATTR_MSG_AUTHENTICATOR)
				{
					msgauthindex = pos;
				}
				entry.addElement(val);
				pos = pos + attrlen;
			}
		}
	}

	public Vector getAttributeValues(int type)
	{
		if (null == attributes)
			initAttributes();
		return (Vector) attributes.get(new Integer(type));
	}

	/**
	 * 
	 * @param type
	 * @return First value under key type if it exists else null
	 */
	public byte[] getAttributeValue(int type)
	{
		Vector vals = getAttributeValues(type);
		if (null != vals)
		{
			return (byte[]) vals.elementAt(0);
		}
		return null;
	}

	/**
	 * 
	 * @param type
	 * @return Concatenated values under key type
	 */
	public byte[] getAttributeConcatValues(int type)
	{
		Vector vals = getAttributeValues(type);
		if (null != vals)
		{
			Enumeration elms = vals.elements();
			int len = 0, start = 0;
			
			while (elms.hasMoreElements()) {
				len = len + ((byte[]) elms.nextElement()).length;
			}
			byte[] values = new byte[len];
			elms = vals.elements();
			while (elms.hasMoreElements()) {
				byte[] src = (byte[]) elms.nextElement();
				System.arraycopy(src, 0, values, start, src.length );
				start = start + src.length;
			}			
			return values;
		}
		return null;
	}
	
	public void setAttributeValue(int type, byte[] value)
	{
		Vector vval = new Vector(1);

		vval.addElement(value);
		if (null == attributes)
			initAttributes();
		attributes.put(new Integer(type), vval);
	}

	private int getInt(int offset)
	{
		return (256 * (0xff & data[offset])) + (0xff & data[offset + 1]);
	}

	/**
	 * Returns the address.
	 * @return InetAddress
	 */
	public String getAddress()
	{
		return address;
	}

	/**
	 * Returns the port.
	 * @return int
	 */
	public int getPort()
	{
		return port;
	}

	public Datagram getDatagram()
	{
		return this.p;
	}

	public byte[] getState()
	{
		return getAttributeValue(RADIUS_ATTR_STATE);
	}

	public void setState(byte[] state)
	{
		setAttributeValue(RADIUS_ATTR_STATE, state);
	}

	/**
	 * Make attributes buffer from attribute table
	 * @param attributes
	 */
	public static byte[] makeAttrBuffer(Hashtable attributes)
	{
		int totallength = 0;
		Enumeration keys = attributes.keys();

		while (keys.hasMoreElements())
		{
			Integer i = (Integer) keys.nextElement();
			Vector vals = (Vector) attributes.get(i);
			for (int j = 0; j < vals.size(); ++j)
			{
				byte[] val = (byte[]) vals.elementAt(j);
				totallength = totallength + 2 + val.length;
			}
		}
		byte[] buffer = new byte[totallength];
		keys = attributes.keys();
		int pos = 0;
		while (keys.hasMoreElements())
		{
			Integer i = (Integer) keys.nextElement();
			Vector vals = (Vector) attributes.get(i);

			for (int j = 0; j < vals.size(); ++j)
			{
				byte[] val = (byte[]) vals.elementAt(j);
				buffer[pos] = (byte) (0xff & i.intValue()); // attr type
				buffer[pos + 1] = (byte) (2 + (0xff & val.length)); // total record length
				System.arraycopy(val, 0, buffer, pos + 2, val.length);
				pos = pos + 2 + val.length;
			}
		}
		return buffer;
	}

	/**
	 * The message is an access request. Compute the message authenticator. 
	 * See RFC 2869
	 * @param secret
	 * @return Autheticator (16 bytes)
	 */
	public byte[] computeRequestAuthenticator(byte[] secret)
	{
		byte[] hash = new byte[16];
		// Note that data.length is _not_ the length of the radius packet  - it is the length of the 
		// datagram buffer -- this cost me some grey hair.
		// More grey hairs - the authenticator was computed across the packet withe attribute MSG-AUTHTICATOR
		// present but set to all zeros -- we have to set it back to 0 here to conpte the packet authenticator.
		// Don't bother putting it back becuase it is never needed again.

		if (msgauthindex == 0)
		{
			return null; // a big error 
		}
		else
		{
			for (int i = 0; i < 16; ++i)
				data[msgauthindex + 2 + i] = 0;
		}
		new OpenSSL().hmacmd5(secret, secret.length, data, getLength(), hash);
		return hash;
	}

	/**
	 * @return
	 */
	public byte[] getUser()
	{
		return getAttributeValue(RADIUS_ATTR_USER_NAME);
	}

	public String getUserMac()
	{
		return new String(getAttributeValue(RADIUS_ATTR_CALLING_STATION_ID)).trim();
	}

	/**
	 * @return
	 */
	public byte[] getRequestAuthenticator()
	{
		byte[] auth = new byte[16];

		System.arraycopy(data, 4, auth, 0, 16);
		return auth;
	}
}
